#pragma once

#include "../core/exception.h"

#include <memory>
#include <cmath>

namespace csg
{
	/* bounding box struct */
	struct bbox_s
	{
		double min_x;
		double max_x;
		double min_y;
		double max_y;
	};

	/* base csg class */
	class base
	{
	public:
		virtual bool is_inside(double x, double y) const = 0;
		virtual std::shared_ptr<csg::base> clone(void) const = 0;
		virtual struct bbox_s get_bounding_box(void) const = 0;

		virtual bool set_param(const std::string& name, const std::string& value) = 0;
	};

	/* rect class */
	class rect : public base
	{
	public:

		// ctor
		rect(double cx, double cy, double width, double height)
		{
			this->m_cx = cx;
			this->m_cy = cy;
			this->m_width = width;
			this->m_height = height;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (_strcmpi(name.c_str(), "x") == 0)
				this->m_cx = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "y") == 0)
				this->m_cy = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "width") == 0)
				this->m_width = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "height") == 0)
				this->m_height = std::atof(value.c_str());
			else
				return false;

			return true;
		}

		// clone class
		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			return std::make_shared<rect>(this->m_cx, this->m_cy, this->m_width, this->m_height);
		}

		// check if point is inside rect
		virtual bool is_inside(double x, double y) const override
		{
			return (std::fabs(x - this->m_cx) <= 0.5 * std::fabs(this->m_width)) && (std::fabs(y - this->m_cy) <= 0.5 * std::fabs(this->m_height));
		}

		// retrieve AABB
		virtual struct bbox_s get_bounding_box(void) const override
		{
			struct bbox_s s;

			s.min_x = std::min(this->m_cx - 0.5 * this->m_width, this->m_cx + 0.5 * this->m_width);
			s.max_x = std::max(this->m_cx - 0.5 * this->m_width, this->m_cx + 0.5 * this->m_width);

			s.min_y = std::min(this->m_cy - 0.5 * this->m_height, this->m_cy + 0.5 * this->m_height);
			s.max_y = std::max(this->m_cy - 0.5 * this->m_height, this->m_cy + 0.5 * this->m_height);

			return s;
		}

	private:
		double m_cx, m_cy, m_width, m_height;
	};

	/* disk class */
	class disk : public base
	{
	public:

		// ctor
		disk(double cx, double cy, double radius)
		{
			this->m_cx = cx;
			this->m_cy = cy;
			this->m_radius = radius;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (_strcmpi(name.c_str(), "x") == 0)
				this->m_cx = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "y") == 0)
				this->m_cy = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "radius") == 0)
				this->m_radius = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "diam") == 0)
				this->m_radius = 0.5 * std::atof(value.c_str());
			else
				return false;

			return true;
		}

		// clone class
		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			return std::make_shared<disk>(this->m_cx, this->m_cy, this->m_radius);
		}

		// check if point is inside rect
		virtual bool is_inside(double x, double y) const override
		{
			return std::hypot(x - this->m_cx, y - this->m_cy) <= std::fabs(this->m_radius);
		}

		// retrieve AABB
		virtual struct bbox_s get_bounding_box(void) const override
		{
			struct bbox_s s;

			s.min_x = std::min(this->m_cx - this->m_radius, this->m_cx + this->m_radius);
			s.max_x = std::max(this->m_cx - this->m_radius, this->m_cx + this->m_radius);

			s.min_y = std::min(this->m_cy - this->m_radius, this->m_cy + this->m_radius);
			s.max_y = std::max(this->m_cy - this->m_radius, this->m_cy + this->m_radius);

			return s;
		}

	private:
		double m_cx, m_cy, m_radius;
	};

	/* move class */
	class move : public base
	{
	public:
		move(std::shared_ptr<base> op, const double x, const double y)
		{
			this->m_op = op;
			this->m_x = x;
			this->m_y = y;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (_strcmpi(name.c_str(), "x") == 0)
				this->m_x = std::atof(value.c_str());
			else if (_strcmpi(name.c_str(), "y") == 0)
				this->m_y = std::atof(value.c_str());
			else if (str_begins_with(name, "op."))
			{
				if (this->m_op == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op->set_param(name.substr(3), value);
			}
			else
				return false;

			return true;
		}

		virtual bool is_inside(double x, double y) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op->is_inside(x - this->m_x, y - this->m_y);
		}

		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return std::make_shared<move>(this->m_op->clone(), this->m_x, this->m_y);
		}

		virtual struct bbox_s get_bounding_box(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			struct bbox_s ret = this->m_op->get_bounding_box();

			ret.min_x -= this->m_x;
			ret.min_y -= this->m_y;
			ret.max_x -= this->m_x;
			ret.max_y -= this->m_y;

			return ret;
		}

	private:
		std::shared_ptr<base> m_op;
		double m_x, m_y;
	};

	/* scale class */
	class scale : public base
	{
	public:
		scale(std::shared_ptr<base> op, const double val)
		{
			this->m_op = op;
			this->m_val = val;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (_strcmpi(name.c_str(), "scale") == 0)
				this->m_val = std::atof(value.c_str());
			else if (str_begins_with(name, "op."))
			{
				if (this->m_op == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op->set_param(name.substr(3), value);
			}
			else
				return false;

			return true;
		}

		virtual bool is_inside(double x, double y) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op->is_inside(x / this->m_val, y / this->m_val);
		}

		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return std::make_shared<scale>(this->m_op->clone(), this->m_val);
		}

		virtual struct bbox_s get_bounding_box(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			struct bbox_s tmp = this->m_op->get_bounding_box();

			auto cx = 0.5 * (tmp.min_x + tmp.max_x);
			auto width = this->m_val * (tmp.max_x - tmp.min_x);

			auto cy = 0.5 * (tmp.min_y + tmp.max_y);
			auto height = this->m_val * (tmp.max_y - tmp.min_y);

			struct bbox_s ret;

			ret.min_x = cx - 0.5 * width;
			ret.max_x = cx + 0.5 * width;
			ret.min_y = cy - 0.5 * height;
			ret.max_y = cy + 0.5 * height;

			return ret;
		}

	private:
		std::shared_ptr<base> m_op;
		double m_val;
	};

	/* bit_not class */
	class bit_not : public base
	{
	public:
		bit_not(std::shared_ptr<base> op)
		{
			this->m_op = op;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op->set_param(name, value);
		}

		virtual bool is_inside(double x, double y) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return !this->m_op->is_inside(x, y);
		}

		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return std::make_shared<bit_not>(this->m_op->clone());
		}

		virtual struct bbox_s get_bounding_box(void) const override
		{
			if (this->m_op == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op->get_bounding_box();
		}

	private:
		std::shared_ptr<base> m_op;
	};

	/* bit_and class */
	class bit_and : public base
	{
	public:
		bit_and(std::shared_ptr<base> op1, std::shared_ptr<base> op2)
		{
			this->m_op1 = op1;
			this->m_op2 = op2;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (str_begins_with(name, "op1."))
			{
				if (this->m_op1 == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op1->set_param(name.substr(4), value);
			}
			else if (str_begins_with(name, "op2."))
			{
				if (this->m_op2 == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op2->set_param(name.substr(4), value);
			}
			else
				return false;

			return true;
		}

		virtual bool is_inside(double x, double y) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op1->is_inside(x, y) && this->m_op2->is_inside(x, y);
		}

		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			return std::make_shared<bit_and>(this->m_op1->clone(), this->m_op2->clone());
		}

		virtual struct bbox_s get_bounding_box(void) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			struct bbox_s op1 = this->m_op1->get_bounding_box();
			struct bbox_s op2 = this->m_op2->get_bounding_box();

			struct bbox_s ret;

			ret.min_x = std::min(op1.min_x, op2.min_x);
			ret.min_y = std::min(op1.min_y, op2.min_y);
			ret.max_x = std::max(op1.max_x, op2.max_x);
			ret.max_y = std::max(op1.max_y, op2.max_y);

			return ret;
		}

	private:
		std::shared_ptr<base> m_op1, m_op2;
	};

	/* bit_or class */
	class bit_or : public base
	{
	public:
		bit_or(std::shared_ptr<base> op1, std::shared_ptr<base> op2)
		{
			this->m_op1 = op1;
			this->m_op2 = op2;
		}

		// set parameters
		virtual bool set_param(const std::string& name, const std::string& value) override
		{
			if (str_begins_with(name, "op1."))
			{
				if (this->m_op1 == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op1->set_param(name.substr(4), value);
			}
			else if (str_begins_with(name, "op2."))
			{
				if (this->m_op2 == nullptr)
					throw_exception(null_pointer_exception);

				return this->m_op2->set_param(name.substr(4), value);
			}
			else
				return false;

			return true;
		}

		virtual bool is_inside(double x, double y) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			return this->m_op1->is_inside(x, y) || this->m_op2->is_inside(x, y);
		}

		virtual std::shared_ptr<csg::base> clone(void) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			return std::make_shared<bit_or>(this->m_op1->clone(), this->m_op2->clone());
		}

		virtual struct bbox_s get_bounding_box(void) const override
		{
			if (this->m_op1 == nullptr || this->m_op2 == nullptr)
				throw_exception(null_pointer_exception);

			struct bbox_s op1 = this->m_op1->get_bounding_box();
			struct bbox_s op2 = this->m_op2->get_bounding_box();

			struct bbox_s ret;

			ret.min_x = std::min(op1.min_x, op2.min_x);
			ret.min_y = std::min(op1.min_y, op2.min_y);
			ret.max_x = std::max(op1.max_x, op2.max_x);
			ret.max_y = std::max(op1.max_y, op2.max_y);

			return ret;
		}

	private:
		std::shared_ptr<base> m_op1, m_op2;
	};
};