#include "../core/improc/map2d.h"
#include "../core/improc/bitmap.h"
#include "../core/improc/morphops.h"
#include "../core/improc/bool.h"

#include "../common/common.h"

#include "stack.h"
#include "csg.h"

enum class edge_mode
{
	interior,
	exterior,
	centered,
};

/* render command */
class render : public textcommand
{
public:
	IMPLEMENT_STORAGE;

	render(void)
	{
		this->m_dpi = 0;
	}

	render(const std::vector<std::string>& args)
	{
		if (args.size() != 2)
			throw_exception(invalid_number_of_arguments_exception);

		this->m_filename = args[0];
		this->m_dpi = atof(args[1].c_str());
	}

	// converts mm <--> inches
	static double _mm2in(const double mm)
	{
		return mm / 25.4;
	};

	static double _in2mm(const double mm)
	{
		return mm * 25.4;
	};

	// misc
	static double _mod(const double val, const double m)
	{
		return val - m * std::floor(val / m);
	};

	// create disk
	static map2d<bool> _disk(const double t, const double dpi)
	{
		auto v = _mm2in(t) * dpi;
		auto s = (size_t)std::ceil(v);

		if ((s % 2) == 0)
			s++;

		map2d<bool> ret(s, s);

		for (size_t y = 0; y < s; y++)
			for (size_t x = 0; x < s; x++)
				ret(x, y) = std::hypot((double)x - 0.5 * (double)s, (double)y - 0.5 * (double)s) <= 0.5 * v;

		return ret;
	};

	static map2d<unsigned char> _render(std::shared_ptr<const csg::base> p, const double left, const double top, const double right, const double bottom, const double dpi, const double hash_period, const double hash_thickness, const double edge_thickness, const edge_mode edge_type)
	{
		// retrieve width
		auto bbox_width = right - left;
		auto bbox_height = bottom - top;

		// create binary image
		auto im_width = (size_t)std::ceil(_mm2in(bbox_width) * dpi);

		while (im_width % 4 != 0)
			im_width++;

		auto im_height = (size_t)std::ceil(_mm2in(bbox_height) * dpi);

		while (im_height % 4 != 0)
			im_height++;

		map2d<bool> m(im_width, im_height);

		for (size_t y = 0; y < m.height(); y++)
			for (size_t x = 0; x < m.width(); x++)
			{
				double _x = left + _in2mm((double)x / dpi);
				double _y = top + _in2mm((double)y / dpi);

				m(x, y) = p->is_inside(_x, _y);
			}

		// compute edges
		map2d<bool> edges;

		switch (edge_type)
		{
		case edge_mode::centered:
			edges = imdilate(m, _disk(0.5 * edge_thickness, dpi)) ^ imerode(m, _disk(0.5 * edge_thickness, dpi));
			break;

		case edge_mode::interior:
			edges = m ^ imerode(m, _disk(edge_thickness, dpi));
			break;

		case edge_mode::exterior:
			edges = m ^ imdilate(m, _disk(edge_thickness, dpi));
			break;
		};

		// create bitmap image
		map2d<unsigned char> bitmap(m.width(), m.height());

		for (size_t y = 0; y < m.height(); y++)
			for (size_t x = 0; x < m.width(); x++)
			{
				double _x = left + _in2mm((double)x / dpi);
				double _y = top + _in2mm((double)y / dpi);

				// compute hash pattern
				bool hash = _mod((_x + _y) * std::sqrt(2) * 0.5, hash_period) < hash_thickness;

				// assign value
				auto black_pixel = (m(x, y) & hash) | edges(x, y);

				bitmap(x, y) = black_pixel ? 0 : 255;
			}

		return bitmap;
	}

	virtual void run(void) const override
	{
		// consts
		const size_t upscale_factor = 3;
		const double bbox_expand = 1.25;
		const double hash_period = 1;
		const double hash_thickness = 0.1;
		const double edge_thickness = 2 * hash_thickness;

		const edge_mode edge_type = edge_mode::centered;

		// get top element
		auto p = stack_top();

		if (p == nullptr)
			return;

		// get boundaries
		auto bbox = p->get_bounding_box();

		auto bbox_cx = 0.5 * (bbox.min_x + bbox.max_x);
		auto bbox_cy = 0.5 * (bbox.min_y + bbox.max_y);

		auto bbox_width = bbox_expand * (bbox.max_x - bbox.min_x);
		auto bbox_height = bbox_expand * (bbox.max_y - bbox.min_y);

		auto left = bbox_cx - 0.5 * bbox_width;
		auto right = bbox_cx + 0.5 * bbox_width;

		auto top = bbox_cy - 0.5 * bbox_height;
		auto bottom = bbox_cy + 0.5 * bbox_height;

		// compute data
		auto upscaled_bitmap = _render(p, left, top, right, bottom, (double)upscale_factor * this->m_dpi, hash_period, hash_thickness, edge_thickness, edge_type);

		// downscale bitmap
		map2d<unsigned char> bitmap(upscaled_bitmap.width() / upscale_factor, upscaled_bitmap.height() / upscale_factor);

		for (size_t y = 0; y < bitmap.height(); y++)
			for (size_t x = 0; x < bitmap.width(); x++)
			{
				double sum = 0;

				for (size_t y2 = 0; y2 < upscale_factor; y2++)
					for (size_t x2 = 0; x2 < upscale_factor; x2++)
						sum += (double)upscaled_bitmap(x * upscale_factor + x2, y * upscale_factor + y2);

				sum /= (double)(upscale_factor * upscale_factor);

				bitmap(x, y) = (unsigned char)std::round(sum);
			}

		// save bitmap file
		try
		{
			save_bitmap(bitmap, this->m_filename, this->m_dpi);
		} catch(...) {}
	}

	virtual std::string log(void) const override
	{
		return get_instance<library::manager>()->get_name() + "/" + format("render \"%s\" %f", this->m_filename.c_str(), this->m_dpi);
	}

private:
	double m_dpi;
	storage::ext::string m_filename;
};

BEGIN_STORAGE(render)
{
	DECLARE_STORAGE(render, m_dpi),
	DECLARE_STORAGE(render, m_filename),
} END_STORAGE(render, version_command_t);

DECLARE_INSTANCEABLE_OBJECT(render, version_command_t);

DECLARE_COMMAND_EX(render, "render");
