#include <stdio.h>
#include <filesystem>

#include "../core/utils/utils.h"
#include "../core/debug/evemon.h"
#include "../core/debug/exdebug.h"
#include "../core/debug/exception.h"
#include "../core/version/node.h"
#include "../core/version/command.h"
#include "../core/version/version.h"
#include "../core/exception.h"

#include "../common/common.h"

#include "graphml.h"

#include <list>
#include <memory>
#include <string>

#ifndef WIN32
#include <unistd.h>
#include <pwd.h>
#endif

#ifdef WIN32
#define EOL "\r\n"
#else
#define EOL "\n"
#endif

class commit_node : public graphml::shape
{
public:
	commit_node(const std::string& label)
	{
		this->m_label = label;
	}

	virtual std::list<std::string> to_string(void) const override
	{
		std::list<std::string> ret;

		ret.push_back("          <y:Shape type=\"roundrectangle\"/>");
		ret.push_back("          <y:Fill color=\"#e0e0e0\" transparent=\"false\"/>");
		ret.push_back("          <y:BorderStyle color=\"#000000\" type=\"line\" width=\"1.0\"/>");
		ret.push_back(format("          <y:NodeLabel>%s</y:NodeLabel>", this->m_label.c_str()));

		return ret;
	}

private:
	std::string m_label;
};

class tag_node : public graphml::shape
{
public:
	tag_node(const std::string& label)
	{
		this->m_label = label;
	}

	virtual std::list<std::string> to_string(void) const override
	{
		std::list<std::string> ret;

		ret.push_back("          <y:Shape type=\"rectangle\"/>");
		ret.push_back("          <y:Fill color=\"#ffffff\" transparent=\"false\"/>");
		ret.push_back("          <y:DropShadow color=\"#808080\" offsetX=\"3\" offsetY=\"3\"/>");
		ret.push_back("          <y:BorderStyle color=\"#000000\" type=\"line\" width=\"2.0\"/>");
		ret.push_back(format("          <y:NodeLabel>%s</y:NodeLabel>", this->m_label.c_str()));

		return ret;
	}

private:
	std::string m_label;
};

class head_node : public graphml::shape
{
public:
	head_node(const std::string& label)
	{
		this->m_label = label;
	}

	virtual std::list<std::string> to_string(void) const override
	{
		std::list<std::string> ret;

		ret.push_back("          <y:Shape type=\"rectangle\"/>");
		ret.push_back("          <y:Fill hasColor=\"false\" transparent=\"false\"/>");
		ret.push_back("          <y:BorderStyle hasColor=\"false\" type=\"line\" width=\"1.0\"/>");
		ret.push_back(format("          <y:NodeLabel>%s</y:NodeLabel>", this->m_label.c_str()));

		return ret;
	}

private:
	std::string m_label;
};

class simple_edge : public graphml::edge
{
public:
	virtual std::list<std::string> to_string(void) const override
	{
		std::list<std::string> ret;

		ret.push_back("          <y:Arrows source=\"none\" target=\"standard\" />");
		ret.push_back("          <y:LineStyle color=\"#000000\" type=\"line\" width=\"1.0\" />");

		return ret;
	}
};

class named_edge : public graphml::edge
{
public:
	named_edge(const std::string& label)
	{
		this->m_label = label;
	}

	virtual std::list<std::string> to_string(void) const override
	{
		std::list<std::string> ret;

		ret.push_back("          <y:Arrows source=\"none\" target=\"standard\" />");
		ret.push_back("          <y:LineStyle color=\"#000000\" type=\"line\" width=\"1.0\" />");
		ret.push_back(format("          <y:EdgeLabel backgroundColor=\"#ffffff\" configuration=\"AutoFlippingLabel\">%s<y:LabelModel><y:SmartEdgeLabelModel autoRotationEnabled=\"true\" defaultAngle=\"0.0\" defaultDistance=\"5.0\"/></y:LabelModel><y:ModelParameter><y:SmartEdgeLabelModelParameter angle=\"0.0\" distance=\"5.0\" distanceToCenter=\"true\" position=\"right\" ratio=\"0.5\" segment=\"0\"/></y:ModelParameter><y:PreferredPlacementDescriptor angle=\"0\" angleOffsetOnRightSide=\"0\" angleReference=\"absolute\" angleRotationOnRightSide=\"co\" distance=\"-1.0\" placement=\"source\" side=\"anywhere\" sideReference=\"relative_to_edge_flow\"/></y:EdgeLabel>", this->m_label.c_str()));

		return ret;
	}

private:
	std::string m_label;
};

// create graphml from project
auto create_graphml(const std::string& project_name, bool show_content=true)
{
	graphml::file obj;
	std::unordered_map<std::string, std::string> table;

	// browse all commits
	auto commits = version::list_commits(project_name);

	for (auto& v : commits)
	{
		auto label = version::get_short_name(project_name, v);

		if(show_content)
		{
			version::proxy<version::node> curr(version::id2path(project_name, v));

			label = "(" + label + ")";

			for (auto& q : curr->get_commands().log())
			{
				label += EOL;
				label += q;
			}
		}

		auto id = obj.add(std::make_shared<commit_node>(label));
		table.emplace(std::make_pair(v, id));
	}

	// link all commits
	for (auto& v : commits)
	{
		version::proxy<version::node> curr(version::id2path(project_name, v));

		auto parents = curr->get_parents();

		if (parents.size() == 0)
		{ }
		else if(parents.size() == 1)
			obj.add(table[version::path2id(parents.front())], table[v], std::make_shared<simple_edge>());
		else
		{
			int i = 1;

			for (auto& p : parents)
				obj.add(table[version::path2id(p)], table[v], std::make_shared<named_edge>(format("%d", i++)));
		}
	}

	// browse all tags
	auto tags = version::list_tags(project_name);

	for (auto& v : tags)
	{
		auto id = obj.add(std::make_shared<tag_node>(v));
		table.emplace(std::make_pair(v, id));
	}

	// link all tags
	for (auto& v : tags)
	{
		auto commit_id = version::resolve_tag(project_name, v);

		obj.add(table[commit_id], table[v], std::make_shared<simple_edge>());
	}

	// browse all heads
	auto heads = version::list_heads(project_name);

	for (auto& v : heads)
	{
		auto id = obj.add(std::make_shared<head_node>("HEAD/" + v));
		table.emplace(std::make_pair(v, id));
	}

	// link all heads
	for (auto& v : heads)
	{
		auto commit_id = version::resolve_head(project_name, v);

		obj.add(table[commit_id], table[v], std::make_shared<simple_edge>());
	}

	return obj;
}

// apply command
bool apply_cmd(std::shared_ptr<version::node>& node, const std::string& cmd)
{
	// stop if node is null
	if (node == nullptr)
		return true;

	// stop on quit/exit/end
	if (_strcmpi(cmd.c_str(), "quit") == 0 || _strcmpi(cmd.c_str(), "exit") == 0 || _strcmpi(cmd.c_str(), "end") == 0)
		return true;

	// undo
	if (_strcmpi(cmd.c_str(), "undo") == 0)
	{
		// pop current transaction list
		if(node->has_transactions())
			node->pop();
		// recurse parents
		else
		{
			// recurse until node has transactions
			while (!node->has_transactions())
			{
				auto parents = node->get_parents();

				if (parents.size() == 0)
					throw_exception(no_parent_exception, "", 1);

				// always use first parent
				version::proxy<version::node> program(parents.front());

				node = program->clone();
			}

			// pop transaction
			node->pop();
		}

		return false;
	}

	// tokenize string
	auto tokens = tokenize(cmd, ' ');

	if (tokens.size() == 0)
		return false;

	// extract name from tokens
	auto cmd_name = tokens[0];
	tokens.erase(tokens.begin());
	
	// create instance
	auto p = create_instance<textcommand, const std::vector<std::string>&>(cmd_name, tokens);

	if (p == nullptr)
		return false;

	node->push(p);

	return false;
}

// read from stdin
void read_stdin(std::shared_ptr<version::node>& node)
{
	while (!feof(stdin))
	{
		char buffer[2048];
		memset(buffer, 0, sizeof(buffer));

		fgets(buffer, sizeof(buffer), stdin);

		if (strlen(buffer) == 0)
			continue;

		if (buffer[strlen(buffer) - 1] == '\n')
			buffer[strlen(buffer) - 1] = '\0';

		if (buffer[strlen(buffer) - 1] == '\r')
			buffer[strlen(buffer) - 1] = '\0';

		if (strlen(buffer) == 0)
			continue;

		if (apply_cmd(node, buffer))
			break;
	}
}

// show help
void show_help(const std::string& name)
{
	printf("- %s help" EOL, name.c_str());
	printf("\tdisplay this manual" EOL);
	printf(EOL);

	printf("- %s init" EOL, name.c_str());
	printf("\tcreate new project" EOL);
	printf(EOL);

	printf("- %s checkout <transaction>" EOL, name.c_str());
	printf("\tcheckout given transaction." EOL);
	printf(EOL);

	printf("- %s resolve <transaction>" EOL, name.c_str());
	printf("\tresolve transaction." EOL);
	printf(EOL);

	printf("- %s tag [<name>]" EOL, name.c_str());
	printf("\tdisplay transaction id of given tag or list all active tags if no transaction is provided." EOL);
	printf(EOL);

	printf("- %s commit [<command>]" EOL, name.c_str());
	printf("\tcommit command on top of HEAD to a new transaction." EOL);
	printf("\tread from stdin if no commands are provided." EOL);
	printf(EOL);

	printf("- %s amend [<command>]" EOL, name.c_str());
	printf("\tamend commands to HEAD to a new transaction." EOL);
	printf("\tread from stdin if no commands are provided." EOL);
	printf(EOL);

	printf("- %s merge <transaction 1> <transaction 2> ..." EOL, name.c_str());
	printf("\tmerge all transactions into a new one." EOL);
	printf(EOL);

	printf("- %s repeat <transaction 1> <transaction 2> ..." EOL, name.c_str());
	printf("\trepeat transactions after HEAD." EOL);
	printf(EOL);

	printf("- %s list-users" EOL, name.c_str());
	printf("\tdisplay all current users working on the project." EOL);
	printf(EOL);

	printf("- %s log [<transaction>]" EOL, name.c_str());
	printf("\tdisplay a log of all commands in the given transaction or from HEAD if no transaction is given." EOL);
	printf(EOL);

	printf("- %s recursive-log [<transaction>]" EOL, name.c_str());
	printf("\tdisplay a log of all commands in the given transaction and above or from HEAD if no transaction is given." EOL);
	printf(EOL);

	printf("- %s export-tree [<filename>]" EOL, name.c_str());
	printf("\tExport .graphml tree structure of project." EOL);
	printf("\tExport to stdout if no filename is provided." EOL);
	printf(EOL);

	printf("- %s run [<transaction>]" EOL, name.c_str());
	printf("\trun all commands in the given transaction or from HEAD if no transaction is given." EOL);
	printf(EOL);
}

// main entry point
int main(int argc, char** argv)
{
	try
	{
		// get program name
		std::string program_name;

		if (argc == 0)
			program_name = "yag";
		else
			program_name = split_file_parts(argv[0]).file;

		// project name is derived from exe name
		std::string project_name = "." + program_name;

		// load user name
#ifdef WIN32

		char buffer[256 + 1];
		DWORD buffer_len = sizeof(buffer);

		GetUserNameA(buffer, &buffer_len);

		std::string user_name(buffer);

#else

		auto puid = getpwuid(geteuid());

		if(__ISNULLPTR(puid))
			throw_exception(null_pointer_exception);

		std::string user_name(puid->pw_name);

#endif

		_debug("connected as \"%s\"", user_name.c_str());

		// load libraries
		for (auto const& curr_entry : std::filesystem::directory_iterator("./"))
		{
			if (!curr_entry.is_regular_file())
				continue;

			auto filename = curr_entry.path().string();

#if _WIN32
			if (_strcmpi(split_file_parts(filename).ext.c_str(), "dll") != 0)
				continue;
#else
			if (_strcmpi(split_file_parts(filename).ext.c_str(), "so") != 0)
				continue;
#endif

			get_instance<library::manager>()->load(filename);
		}

		// help
		if (argc == 2 && _strcmpi(argv[1], "help") == 0)
			show_help(program_name);
		// init
		else if (argc == 2 && _strcmpi(argv[1], "init") == 0)
		{
			version::init(project_name, user_name);

			// create root node
			auto root_id = version::save(project_name, version::create_node(user_name, {}));

			// set as head
			version::set_head(project_name, user_name, root_id);

			printf("root node is %s" EOL, version::get_short_name(project_name, root_id).c_str());
		}
		// checkout
		else if (argc == 3 && _strcmpi(argv[1], "checkout") == 0)
		{
			auto filename = version::checkout(project_name, user_name, argv[2]);

			printf("transaction %s checked out" EOL, version::get_short_name(project_name, filename).c_str());
		}
		// tag
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "tag") == 0)
		{
			if (argc == 2)
			{
				auto list = version::list_tags(project_name);

				if (list.size() == 0)
					printf("There are no tags in this project." EOL);
				else
				{
					printf("Tags associated to this project:" EOL);

					for (auto& v : list)
						printf("\t- %s" EOL, v.c_str());
				}
			}
			else
			{
				auto head_id = version::resolve_head(project_name, user_name);

				version::set_tag(project_name, argv[2], head_id);

				printf("transaction %s assigned as \"%s\"\n", version::get_short_name(project_name, head_id).c_str(), argv[2]);
			}
		}
		// peek
		else if (argc == 2 && _strcmpi(argv[1], "list-users") == 0)
		{
			auto list = version::list_heads(project_name);

			if (list.size() == 0)
				printf("There are no active user in this project." EOL);
			else
			{
				printf("Users associated to this project:" EOL);

				for (auto& v : list)
					printf("\t- %s" EOL, v.c_str());
			}
		}
		// recursive-log
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "recursive-log") == 0)
		{
			std::string commit_id;

			if (argc == 2)
				commit_id = version::resolve_head(project_name, user_name);
			else
				commit_id = version::resolve_commit_id(project_name, user_name, argv[2]);

			version::proxy<version::node> program(version::id2path(project_name, commit_id));

			auto list = program->log();

			if (list.size() == 0)
				printf("No data" EOL);
			else
			{
				for (auto& v : list)
					printf("%s" EOL, v.c_str());
			}
		}
		// log
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "log") == 0)
		{
			std::string commit_id;

			if (argc == 2)
				commit_id = version::resolve_head(project_name, user_name);
			else
				commit_id = version::resolve_commit_id(project_name, user_name, argv[2]);

			version::proxy<version::node> program(version::id2path(project_name, commit_id));

			auto list = program->get_commands().log();

			if (list.size() == 0)
				printf("No data" EOL);
			else
			{
				for (auto& v : list)
					printf("%s" EOL, v.c_str());
			}
		}
		// run
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "run") == 0)
		{
			std::string commit_id;

			if (argc == 2)
				commit_id = version::resolve_head(project_name, user_name);
			else
				commit_id = version::resolve_commit_id(project_name, user_name, argv[2]);

			version::proxy<version::node> program(version::id2path(project_name, commit_id));

			auto commands = program->integrate();

			commands.run();
		}
		// commit
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "commit") == 0)
		{
			// resolve head
			auto head_id = version::resolve_head(project_name, user_name);
			
			// create new node
			auto node = version::create_node(user_name, { version::id2path(project_name, head_id) });

			// push commands
			if (argc == 3)
				apply_cmd(node, argv[2]);
			else
				read_stdin(node);

			// save node and update head
			auto commit_id = version::save(project_name, node);

			version::set_head(project_name, user_name, commit_id);

			// display
			printf("transaction recorded as %s" EOL, version::get_short_name(project_name, commit_id).c_str());
		}
		// amend commit
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "amend") == 0)
		{
			// resolve head
			auto head_id = version::resolve_head(project_name, user_name);

			// clone node
			version::proxy<version::node> program(version::id2path(project_name, head_id));

			auto node = program->clone();

			// push commands
			if (argc == 3)
				apply_cmd(node, argv[2]);
			else
				read_stdin(node);

			// save node and update head
			auto commit_id = version::save(project_name, node);

			version::set_head(project_name, user_name, commit_id);

			// display
			printf("transaction recorded as %s" EOL, version::get_short_name(project_name, commit_id).c_str());
		}
		// merge
		else if (argc >= 3 && _strcmpi(argv[1], "merge") == 0)
		{
			std::list<std::string> parents;

			std::string list = "";

			for (size_t i = 2; i < argc; i++)
			{
				auto commit_id = version::resolve_commit_id(project_name, user_name, argv[i]);

				parents.push_back(version::id2path(project_name, commit_id));

				list = list + version::get_short_name(project_name, commit_id);

				if ((i + 2) < argc)
					list += ", ";
				else if ((i + 1) < argc)
					list += " and ";
			}

			// create node
			auto node_id = save(project_name, version::create_node(user_name, parents));

			// set as head
			version::set_head(project_name, user_name, node_id);

			// display
			printf("transactions %s merged as %s" EOL, list.c_str(), version::get_short_name(project_name, node_id).c_str());
		}
		// merge
		else if (argc == 3 && _strcmpi(argv[1], "resolve") == 0)
		{
			printf("%s resolved as %s" EOL, argv[2], version::get_short_name(project_name, version::resolve_commit_id(project_name, user_name, argv[2])).c_str());
		}
		// export graphml file
		else if ((argc == 2 || argc == 3) && _strcmpi(argv[1], "export-tree") == 0)
		{
			auto obj = create_graphml(project_name);

			std::string total = "";

			for (auto& v : obj.to_string())
			{
				if (total != "")
					total += EOL;

				total += v;
			}

			if (argc == 2 || _strcmpi(argv[2], "-") == 0)
				printf("%s", total.c_str());
			else
			{
				FILE* p = nullptr;

				fopen_s(&p, version::cleanup(argv[2]).c_str(), "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, argv[2]);

				fputs(total.c_str(), p);

				fclose(p);
			}
		}
		// repeat transactions on head
		else if (argc >= 3 && _strcmpi(argv[1], "repeat") == 0)
		{
			std::string list = "";

			// resolve head
			auto head_id = version::resolve_head(project_name, user_name);

			// this will hold the current node
			std::shared_ptr<version::node> node;

			// browse all commits in list
			for (size_t i = 2; i < argc; i++)
			{
				auto commit_id = version::resolve_commit_id(project_name, user_name, argv[i]);

				// load commit
				version::proxy<version::node> program(version::id2path(project_name, commit_id));

				node = program->clone();

				// check number of parents
				if (node->get_parents().size() != 1)
					throw_exception(cannot_repeat_exception, commit_id);

				// replace parent
				node->replace_parents({ version::id2path(project_name, head_id) });

				// save and update head
				head_id = save(project_name, node);

				version::set_head(project_name, user_name, head_id);

				// add short name to list
				list = list + version::get_short_name(project_name, commit_id);

				if ((i + 2) < argc)
					list += ", ";
				else if ((i + 1) < argc)
					list += " and ";
			}

			// display
			printf("transactions %s repeated and stored as %s" EOL, list.c_str(), version::get_short_name(project_name, head_id).c_str());
		}
		// otherelse show help and return error
		else
		{
			show_help(program_name);

			return 1;
		}
						
		// return successfuly
		return 0;
	}
	catch (exception::base& e)
	{
		printf("%s" EOL, e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("%s" EOL, e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!" EOL);

		_critical("unknown error!");
	}

	// signal error
	return 1;
}
